#include <iostream>
#include <exception>
#include <fstream>
#include <sstream>
#include <vector>
#include "ics46goody.hpp"
#include "stopwatch.hpp"

template<class T>
class LN {
  public:
    LN ()                        : next(nullptr){}
    LN (const LN<T>& ln)         : value(ln.value), next(ln.next){}
    LN (T v, LN<T>* n = nullptr) : value(v), next(n){}

    T      value;
    LN<T>* next;
};


int sum(LN<int>* l) {
  int sum = 0;
  for (LN<int>* p = l; p!=nullptr; p=p->next)
    sum += p->value;
  return sum;
}


template<class T>
std::ostream& operator << (std::ostream& outs, LN<T>* l) {
  for (LN<T>* p = l; p!=nullptr; p=p->next)
    std::cout << p->value << "->";
  std::cout << "nullptr";
  return outs;
}


template<class T>
int length (LN<T>* l) {
  int answer = 0;
  for (LN<T>* p = l; p!=nullptr; p=p->next)
    answer++;
  return answer;
}


template<class T>
int count_occurrences (LN<T>* l, T to_count) {
  int answer = 0;
  for (LN<T>* p = l; p!=nullptr; p=p->next)
    if (p->value == to_count)
      ++answer;
  return answer;
}


template<class T>
bool is_sorted (LN<T>* l) {
  if (l == nullptr || l->next == nullptr) //Shortcircuit || is critical here
    return true;
  T prev = l->value;

  for (LN<T>* p = l->next;  p!=nullptr;  prev = p->value, p=p->next) {
    if (prev > p->value)
      return false;
  }
  return true;
}


template<class T>
int is_sorted2 (LN<T>* l) {
  if (l == nullptr || l->next == nullptr) //Shortcircuit || is critical here
    return true;

  for (LN<T>* p = l; p->next != nullptr; p = p->next)
    if (p->value > p->next->value)
      return false;
  return true;
}


template<class T>
LN<T>* add_front(LN<T>* l, T value) { //O(1)
  return new LN<T>(value,l);
}


template<class T>
LN<T>* add_rear(LN<T>* l, T value) {  //O(N)
  if (l == nullptr)
    return new LN<T>(value); //Default argument for 2nd parameter is nullptr

  for (LN<T>* p = l; /*see body*/; p = p->next)
    if (p->next == nullptr) {
      p->next = new LN<T>(value);
      break;
    }

  return l;
}


template<class T>
LN<T>* build_linked_list_simple(T values[], int length) {  //O(N^2)
  if (length == 0)
    return nullptr;

  LN<T>* front = nullptr;
  for (int i=0; i<length; ++i)
    front = add_rear(front,values[i]);

  return front;
}


template<class T>
LN<T>* build_linked_list_fast(T values[], int length) {  //O(N)
  if (length == 0)
    return nullptr;

  LN<T>* front      = new LN<T>(values[0]);
  LN<T>* rear_cache = front;
  for (int i=1; i<length; ++i)
    rear_cache = rear_cache->next = new LN<T>(values[i]);

  return front;
}


template<class T>
LN<T>* find (LN<T>* l, T to_find) {
  for (LN<T>* p = l; p != nullptr; p = p->next)
    if (p->value == to_find)
       return p;

  return nullptr;
}


template<class T>
LN<T>* remove_lookahead (LN<T>* l, T to_remove) {
  if (l == nullptr)
    return nullptr;

  if(l->value == to_remove)
    l = l->next;
  else{
    for (LN<T>* p = l; p->next != nullptr; p = p->next)
      if (p->next->value == to_remove) {
        LN<T>* to_delete = p->next;
        p->next = p->next->next;
        delete to_delete;
        break;
      }
  }

  return l;
}


template<class T>
LN<T>* remove_ghost (LN<T>* l, T to_remove) {
  if (l == nullptr)
    return nullptr;

  if(l->value == to_remove)
    l = l->next;
  else{
    LN<T>* prev = l;
    for (LN<T>* p = l->next; p != nullptr; prev = p, p = p->next)
      if (p->value == to_remove) {
        prev->next = p->next;
        delete p;
        break;
      }
  }

  return l;
}


template<class T>
LN<T>*  copy (LN<T>* l) {
  if (l == nullptr)
    return nullptr;

  LN<T>* front      = new LN<T>(l->value);  //Guaranteed to exist: l != nullptr
  LN<T>* rear_cache = front;
  for (l = l->next; l != nullptr; l = l->next)
    rear_cache = rear_cache->next = new LN<T>(l->value);

  return front;
}


template<class T>
bool equals (LN<T>* l1, LN<T>* l2) {
  for (; l1 != nullptr && l2 != nullptr; l1 = l1->next, l2 = l2->next)
    if (l1->value != l2->value)
      return false;

  return l1 == nullptr && l2 == nullptr;
}


template<class T>
void add_front_reference(LN<T>*& l, T value) {
  l = new LN<T>(value,l);
}


template<class T>
void add_rear_reference(LN<T>*& l, T value) {
  if (l == nullptr)
    l = new LN<T>(value); //Default argument for 2nd parameter is nullptr
  else {
    for (LN<T>* p = l; /*see body for termination*/; p = p->next)
      if (p->next == nullptr) {
        p->next = new LN<T>(value);
        break;
      }
  }
}


template<class T>
void remove_lookahead_reference (LN<T>*& l, T to_remove) {
  if (l == nullptr)
    return;
  else if(l->value == to_remove)
    l = l->next;
  else{
    for (LN<T>* p = l; p->next != nullptr; p = p->next)
      if (p->next->value == to_remove) {
        LN<T>* to_delete = p->next;
        p->next = p->next->next;
        delete to_delete;
        break;
      }
  }
}


template<class T>
void remove_ptp (LN<T>*& l, T to_remove) {
  LN<T>** p = &l;
  for (; (*p) != nullptr; p= &((*p)->next))
    if ((*p)->value == to_remove) {
      LN<T>* to_delete = *p;
      (*p) = (*p)->next;
      delete to_delete;
      break;
    }
}


std::vector<int> prompt_vector(std::string preface, std::string message = "  Enter next element") {
  std::vector<int> values;
  for (;;) {
    int e = ics::prompt_int(preface + message + "(-1 to quit)");
    if (e == -1)
      break;
    values.push_back(e);
  }
  return values;
}





LN<int>* ll;

std::string menu_prompt (std::string preface) {
  std::cout << "\n\n"+preface+"linked list = " << ll << std::endl;
  std::cout << preface+"Mutators                   Accessors                  General" << std::endl;
  std::cout << preface+"  af  - add front             s   - sum                  bs - build... simple" << std::endl;
  std::cout << preface+"  ar  - add rear              l   - length               bf - build... fast"   << std::endl;
  std::cout << preface+"  rl  - remove lookahead      co  - count occurrences    lf - load file"       << std::endl;
  std::cout << preface+"  rg  - remove ghost          is  - is_sorted            st - speed test"      << std::endl;
  std::cout << preface+"  rr  - remove reference      is2 - is_sorted2           q   - quit"           << std::endl;
  std::cout << preface+"  x   - set to nullptr        f   - find"   << std::endl;
  std::cout << preface+"  afr - af reference          cp  - copy"   << std::endl;
  std::cout << preface+"  arr - ar reference          =   - equals" << std::endl;
  std::cout << preface+"  rlr - rl reference "                      << std::endl;
  std::cout << preface+"  rp  - remove ptr to ptr "                 << std::endl;

  std::string allowable[] = {"af","ar","rl","rg","rr","x","afr","arr","rlr","rp","s","l","co","is","is2","f","cp","=","bs","bf","lf","st","q",""};
  return ics::prompt_string("\n"+preface+"Enter linked list command","",allowable);
}

void process_commands(std::string preface) {
  for (;;) try{
    std::string command = menu_prompt(preface);
    //KLUDGE: replace extractions with ics::prompt_int, when written

    if (command == "af")
      ll = add_front(ll,ics::prompt_int(preface+"  Enter element to add at front"));

    else if (command == "ar")
      ll = add_rear(ll,ics::prompt_int(preface+"  Enter element to add at rear"));

    else if (command == "rl")
      ll = remove_lookahead(ll,ics::prompt_int(preface+"  Enter element to remove"));

    else if (command == "rg")
      ll = remove_ghost(ll,ics::prompt_int(preface+"  Enter element to remove"));

    else if (command == "x")
      ll = nullptr;

    else if (command == "afr")
      add_front_reference(ll,ics::prompt_int(preface+"  Enter element to add at front"));

    else if (command == "arr")
      add_rear_reference(ll,ics::prompt_int(preface+"  Enter element to add at rear"));

    else if (command == "rlr")
      remove_lookahead_reference(ll,ics::prompt_int(preface+"  Enter element to remove"));

    else if (command == "rp")
      remove_ptp(ll,ics::prompt_int(preface+"  Enter element to remove"));

    else if (command == "s")
      std::cout << "  sum = " << sum(ll) << std::endl;

    else if (command == "l")
      std::cout << "  length = " << length(ll) << std::endl;

    else if (command == "co") {
      int e = ics::prompt_int(preface+"  Enter element to count");
      std::cout << "  count occurrences = " << count_occurrences(ll,e) << std::endl;
    }

    else if (command == "is")
      std::cout << preface+"  is_sorted = " << is_sorted(ll) << std::endl;

    else if (command == "is2")
      std::cout << preface+"  is_sorted = " << is_sorted2(ll) << std::endl;


    else if (command == "f") {
      LN<int>* f = find(ll,ics::prompt_int(preface+"  Enter element to find"));
      std::cout << preface+"  that (and following) = " << f << std::endl;
    }

    else if (command == "cp") {
      LN<int>* cp = copy(ll);
      std::cout << preface+"  copy = " << cp << std::endl;
    }

    else if (command == "=") {
      std::vector<int> values = prompt_vector(preface);
      LN<int>* t = nullptr;
      for (unsigned i=0; i<values.size(); ++i)
        t = add_rear(t,values[i]);
      std::cout << preface+"  ll = " << ll << " ?? t = " << t << std::endl;
      std::cout << preface+"  equlas(ll,ll) = " << equals(ll,t) << std::endl;
    }

    else if (command == "bs") {
      std::vector<int> values = prompt_vector(preface);
      int* t = new int[values.size()];
      for (unsigned i=0; i<values.size(); ++i)
        t[i] = values[i];
      ll = build_linked_list_simple(t,values.size());
    }

    else if (command == "bf") {
      std::vector<int> values = prompt_vector(preface);
      int* t = new int[values.size()];
      for (unsigned i=0; i<values.size(); ++i)
        t[i] = values[i];
      ll = build_linked_list_fast(t,values.size());
    }

    else if (command == "lf") {
      ll = nullptr;
      std::ifstream ll_file;
      ics::safe_open(ll_file,preface+"  Enter file name to read", "load.txt");
      std::string line;
      while (getline(ll_file,line)) {
        std::istringstream input(line);
        int e;
        input >> e;
        ll = add_rear(ll,e);
      }
    }

    else if (command == "st") {
      int how_many = ics::prompt_int(preface+"  Enter # to test");
      int* t = new int[how_many];
      for (int i=0; i<how_many; ++i)
        t[i] = i;

      ics::Stopwatch s;
      s.start();
      build_linked_list_simple(t,how_many);
      s.stop();
      std::cout << "Time for build...simple = " << s.read() << " seconds" << std::endl;

      s.reset();
      s.start();
      build_linked_list_fast(t,how_many);
      s.stop();
      std::cout << "Time for build...fast = " << s.read() << " seconds" << std::endl;

    }

    else if (command == "q")
      break;

    else
      std::cout << preface+"\""+command+"\" is unknown command" << std::endl;

  } catch (const std::exception& e) {
    std::cout << e.what() << std::endl;
  }

}

int main() {
  process_commands("");
  return 0;
}
